@@ -15,28 +15,64 @@ package IO::Socket::SSL;
 
 use IO::Socket;
 use Net::SSLeay 1.21;
+use Exporter ();
+use Scalar::Util 'dualvar';
+use Errno 'EAGAIN';
 use Carp;
 use strict;
-use vars qw(@ISA $VERSION $DEBUG $ERROR $GLOBAL_CONTEXT_ARGS);
+
+use vars qw(@ISA $VERSION $DEBUG $SSL_ERROR $GLOBAL_CONTEXT_ARGS @EXPORT );
+
+{
+    # These constants will be used in $! at return from SSL_connect, 
+    # SSL_accept, generic_read and write, thus notifying the caller
+    # the usual way of problems. Like with EAGAIN, EINPROGRESS..
+    # these are especially important for non-blocking sockets
+
+    my $x = Net::SSLeay::ERROR_WANT_READ();
+    use constant SSL_WANT_READ  => dualvar( \$x, 'SSL wants a read first' );
+    my $y = Net::SSLeay::ERROR_WANT_WRITE();
+    use constant SSL_WANT_WRITE => dualvar( \$y, 'SSL wants a write first' );
+
+    @EXPORT = qw( SSL_WANT_READ SSL_WANT_WRITE $SSL_ERROR );
+}
 
 BEGIN {
     # Declare @ISA, $VERSION, $GLOBAL_CONTEXT_ARGS
     @ISA = qw(IO::Socket::INET);
-    $VERSION = '0.97';
+    $VERSION = '1.02';
     $GLOBAL_CONTEXT_ARGS = {};
 
     #Make $DEBUG another name for $Net::SSLeay::trace
     *DEBUG = \$Net::SSLeay::trace;
 
+    #Compability
+    *ERROR = \$SSL_ERROR;
+
     # Do Net::SSLeay initialization
     Net::SSLeay::load_error_strings();
     Net::SSLeay::SSLeay_add_ssl_algorithms();
     Net::SSLeay::randomize();
+
 }
 
-sub import { foreach (@_) { @ISA=qw(IO::Socket::INET), next if /inet4/i;
-			    @ISA=qw(IO::Socket::INET6), next if /inet6/i;
-			    $DEBUG=$1 if /debug(\d)/; }}
+# Export some stuff
+# inet4|inet6|debug will be handeled by myself, everything
+# else will be handeld the Exporter way
+sub import { 
+    my $class = shift;
+
+    my @export;
+    foreach (@_) { 
+	@ISA=qw(IO::Socket::INET), next if /inet4/i;
+	@ISA=qw(IO::Socket::INET6), next if /inet6/i;
+	$DEBUG=$1, next if /debug(\d)/; 
+	push @export,$_
+    }
+
+    @_ = ( $class,@export );
+    goto &Exporter::import;
+}
 
 # You might be expecting to find a new() subroutine here, but that is
 # not how IO::Socket::INET works.  All configuration gets performed in
@@ -48,10 +84,25 @@ sub configure {
     my ($self, $arg_hash) = @_;
     return _invalid_object() unless($self);
 
+    # force initial blocking 
+    # otherwise IO::Socket::SSL->new might return undef if the
+    # socket is nonblocking and it fails to connect immediatly
+    # for real nonblocking behavior one should create a nonblocking
+    # socket and later call connect explicitly
+    my $blocking = delete $arg_hash->{Blocking};
+
+    # because Net::HTTPS simple redefines blocking() to {} (e.g
+    # return undef) and IO::Socket::INET does not like this we
+    # set Blocking only explicitly if it was set
+    $arg_hash->{Blocking} = 1 if defined ($blocking);
+
     $self->configure_SSL($arg_hash) || return;
 
-    return ($self->SUPER::configure($arg_hash)
-	|| $self->error("@ISA configuration failed"));
+    $self->SUPER::configure($arg_hash)
+	|| return $self->error("@ISA configuration failed");
+
+    $self->blocking(0) if defined $blocking && !$blocking;
+    return $self;
 }
 
 sub configure_SSL {
@@ -61,8 +112,6 @@ sub configure_SSL {
     my %default_args =
 	('Proto'         => 'tcp',
 	 'SSL_server'    => $is_server,
-	 'SSL_key_file'  => $is_server ? 'certs/server-key.pem'  : 'certs/client-key.pem',
-	 'SSL_cert_file' => $is_server ? 'certs/server-cert.pem' : 'certs/client-cert.pem',
 	 'SSL_ca_file'   => 'certs/my-ca.pem',
 	 'SSL_ca_path'   => 'ca/',
 	 'SSL_use_cert'  => $is_server,
@@ -70,8 +119,18 @@ sub configure_SSL {
 	 'SSL_version'   => 'sslv23',
 	 'SSL_verify_mode' => Net::SSLeay::VERIFY_NONE(),
 	 'SSL_verify_callback' => 0,
-	 'SSL_cipher_list' => 'ALL:!LOW:!EXP');
-
+    );
+     
+    # SSL_key_file and SSL_cert_file will only be set in defaults if 
+    # SSL_key|SSL_key_file resp SSL_cert|SSL_cert_file are not set in
+    # $args_hash
+    foreach my $k (qw( key cert )) {
+	next if exists $arg_hash->{ "SSL_${k}" };
+	next if exists $arg_hash->{ "SSL_${k}_file" };
+    	$default_args{ "SSL_${k}_file" } = $is_server 
+	    ?  "certs/server-${k}.pem" 
+	    :  "certs/client-${k}.pem";
+    }	
 
     #Replace nonexistent entries with defaults
     %$arg_hash = ( %default_args, %$GLOBAL_CONTEXT_ARGS, %$arg_hash );
@@ -86,60 +145,107 @@ sub configure_SSL {
     }
 
     ${*$self}{'_SSL_arguments'} = $arg_hash;
-    ${*$self}{'_SSL_ctx'} = new IO::Socket::SSL::SSL_Context($arg_hash) || return;
+    ${*$self}{'_SSL_ctx'} = IO::Socket::SSL::SSL_Context->new($arg_hash) || return;
     ${*$self}{'_SSL_opened'} = 1 if ($is_server);
 
     return $self;
 }
 
 
+sub _set_rw_error {
+    my ($self,$ssl,$rv) = @_;
+    my $err = Net::SSLeay::get_error($ssl,$rv);
+    $SSL_ERROR = 
+	$err == Net::SSLeay::ERROR_WANT_READ()  ? SSL_WANT_READ :
+	$err == Net::SSLeay::ERROR_WANT_WRITE() ? SSL_WANT_WRITE :
+	return;
+    $! ||= EAGAIN;
+    ${*$self}{'_SSL_last_err'} = $SSL_ERROR if (ref($self));
+    return 1;
+}
+
+
 #Call to connect occurs when a new client socket is made using
 #IO::Socket::INET
 sub connect {
     my $self = shift || return _invalid_object();
+    return $self if ${*$self}{'_SSL_opened'};  # already connected
 
-    my $socket = $self->SUPER::connect(@_)
-	|| return $self->error("@ISA connect attempt failed");
-
-    return $self->connect_SSL($socket) || $self->fatal_ssl_error;
+    if ( ! ${*$self}{'_SSL_opening'} ) {
+	# call SUPER::connect if the underlying socket is not connected
+	# if this fails this might not be an error (e.g. if $! = EINPROGRESS
+	# and socket is nonblocking this is normal), so keep any error
+	# handling to the client
+	#DEBUG( 'socket not yet connected' );
+	$self->SUPER::connect(@_) || return;
+	#DEBUG( 'socket connected' );
+    }
+    return $self->connect_SSL;
 }
 
 
 sub connect_SSL {
-    my ($self, $socket) = @_;
-    my $arg_hash = ${*$self}{'_SSL_arguments'};
-    ${*$self}{'_SSL_opened'}=1;
+    my $self = shift;
 
-    my $fileno = ${*$self}{'_SSL_fileno'} = fileno($socket);
-    return $self->error("Socket has no fileno") unless (defined $fileno);
+    my ($ssl,$ctx);
+    if ( ! ${*$self}{'_SSL_opening'} ) {
+	# start ssl connection
+	#DEBUG( 'ssl handshake not started' );
+	${*$self}{'_SSL_opening'} = 1;
+	my $arg_hash = ${*$self}{'_SSL_arguments'};
 
-    my $ctx = ${*$self}{'_SSL_ctx'};  # Reference to real context
-    my $ssl = ${*$self}{'_SSL_object'} = Net::SSLeay::new($ctx->{context})
-	|| return $self->error("SSL structure creation failed");
+	my $fileno = ${*$self}{'_SSL_fileno'} = fileno($self);
+	return $self->error("Socket has no fileno") unless (defined $fileno);
 
-    Net::SSLeay::set_fd($ssl, $fileno)
-	|| return $self->error("SSL filehandle association failed");
+	$ctx = ${*$self}{'_SSL_ctx'};  # Reference to real context
+	$ssl = ${*$self}{'_SSL_object'} = Net::SSLeay::new($ctx->{context})
+	    || return $self->error("SSL structure creation failed");
 
-    Net::SSLeay::set_cipher_list($ssl, $arg_hash->{'SSL_cipher_list'})
-	|| return $self->error("Failed to set SSL cipher list");
+	Net::SSLeay::set_fd($ssl, $fileno)
+	    || return $self->error("SSL filehandle association failed");
 
-    my ($addr, $port) = @{$arg_hash}{'PeerAddr','PeerPort'};
-    my $session = $ctx->session_cache($addr, $port);
-    Net::SSLeay::set_session($ssl, $session) if ($session);
-
-    while (Net::SSLeay::connect($ssl)<1) {
-	$self->error("SSL connect attempt failed");
-	if ($self->errstr =~ /SSL wants a (write|read) first!/) {
-	    require IO::Select;
-	    my $sel = new IO::Select($socket);
-	    my ($timeout) = grep { defined $_ } (@{$arg_hash}{"Timeout","timeout","io_socket_timeout"});
-	    next if (($1 eq 'write') ? $sel->can_write($timeout) : $sel->can_read($timeout));
+	if ( my $cl = $arg_hash->{SSL_cipher_list} ) {
+	    Net::SSLeay::set_cipher_list($ssl, $cl )
+	    	|| return $self->error("Failed to set SSL cipher list");
 	}
-	return;
+
+	my $session = $ctx->session_cache( $arg_hash->{PeerAddr}, $arg_hash->{PeerPort} );
+	Net::SSLeay::set_session($ssl, $session) if ($session);
     }
 
-    if ($ctx->has_session_cache && !$session) {
-	$ctx->session_cache($addr, $port, Net::SSLeay::get1_session($ssl));
+    $ssl ||= ${*$self}{'_SSL_object'};
+
+    $SSL_ERROR = undef;
+    #DEBUG( 'calling ssleay::connect' );
+    my $rv = Net::SSLeay::connect($ssl);
+    #DEBUG( "rv=$rv" );
+    if ( $rv < 0 ) {
+	unless ( $self->_set_rw_error( $ssl,$rv )) {
+	    $self->error("SSL connect attempt failed with unknown error");
+	    delete ${*$self}{'_SSL_opening'};
+	    ${*$self}{'_SSL_opened'} = 1;
+	    return $self->fatal_ssl_error();
+	}
+	#DEBUG( 'ssl handshake in progress' );
+	return;
+    } elsif ( $rv == 0 ) {
+	delete ${*$self}{'_SSL_opening'};
+	$self->error("SSL connect attempt failed because of handshake problems" );
+	${*$self}{'_SSL_opened'} = 1;
+	return $self->fatal_ssl_error();
+    }
+
+    #DEBUG( 'ssl handshake done' );
+    # ssl connect successful
+    delete ${*$self}{'_SSL_opening'};
+    ${*$self}{'_SSL_opened'}=1;
+
+    $ctx ||= ${*$self}{'_SSL_ctx'};
+    if ( $ctx->has_session_cache ) {
+	my $arg_hash = ${*$self}{'_SSL_arguments'};
+	my ($addr,$port) = ( $arg_hash->{PeerAddr}, $arg_hash->{PeerPort} );
+	my $session = $ctx->session_cache( $addr,$port );
+	$ctx->session_cache( $addr,$port, Net::SSLeay::get1_session($ssl) ) if !$session;
     }
 
     tie *{$self}, "IO::Socket::SSL::SSL_HANDLE", $self;
@@ -153,64 +259,98 @@ sub connect_SSL {
 sub accept {
     my $self = shift || return _invalid_object();
     my $class = shift || 'IO::Socket::SSL';
-    my $arg_hash = ${*$self}{'_SSL_arguments'};
 
-    my ($socket, $peer) = $self->SUPER::accept($class)
-	or return $self->error("@ISA accept failed");
+    my $socket = ${*$self}{'_SSL_opening'};
+    if ( ! $socket ) {
+	# underlying socket not done
+	#DEBUG( 'no socket yet' );
+	$socket = $self->SUPER::accept($class) || return;
+	#DEBUG( 'accept created normal socket '.$socket );
+    }
 
-    $socket->accept_SSL(${*$self}{'_SSL_ctx'}, $arg_hash)
-	|| return($self->error($ERROR) || $socket->fatal_ssl_error);
+    $self->accept_SSL($socket) || return;
+    #DEBUG( 'accept_SSL ok' );
 
-    return wantarray ? ($socket, $peer) : $socket;
+    return wantarray ? ($socket, getpeername($socket) ) : $socket;
 }
 
 sub accept_SSL {
-    my ($socket, $ctx, $arg_hash) = @_;
-    ${*$socket}{'_SSL_arguments'} = { %$arg_hash, SSL_server => 0 };
-    ${*$socket}{'_SSL_ctx'} = $ctx;
-    ${*$socket}{'_SSL_opened'} = 1;
+    my ($self,$socket) = @_;
+    $socket ||= $self;
 
-    my $fileno = ${*$socket}{'_SSL_fileno'} = fileno($socket);
-    return $socket->error("Socket has no fileno") unless (defined $fileno);
+    my $ssl;
+    if ( ! ${*$self}{'_SSL_opening'} ) {
+	#DEBUG( 'starting sslifying' );
+	${*$self}{'_SSL_opening'} = $socket;
+	my $arg_hash = ${*$self}{'_SSL_arguments'};
+	${*$socket}{'_SSL_arguments'} = { %$arg_hash, SSL_server => 0 };
+	my $ctx = ${*$socket}{'_SSL_ctx'} = ${*$self}{'_SSL_ctx'};
 
-    my $ssl = ${*$socket}{'_SSL_object'} = Net::SSLeay::new($ctx->{context})
-	|| return $socket->error("SSL structure creation failed");
+	my $fileno = ${*$socket}{'_SSL_fileno'} = fileno($socket);
+	return $socket->error("Socket has no fileno") unless (defined $fileno);
 
-    Net::SSLeay::set_fd($ssl, $fileno)
-	|| return $socket->error("SSL filehandle association failed");
+	$ssl = ${*$socket}{'_SSL_object'} = Net::SSLeay::new($ctx->{context})
+	    || return $socket->error("SSL structure creation failed");
 
-    Net::SSLeay::set_cipher_list($ssl, $arg_hash->{'SSL_cipher_list'})
-	|| return $socket->error("Failed to set SSL cipher list");
+	Net::SSLeay::set_fd($ssl, $fileno)
+	    || return $socket->error("SSL filehandle association failed");
 
-    while (Net::SSLeay::accept($ssl)<1) {
-	$socket->error("SSL accept attempt failed");
-	if ($socket->errstr =~ /SSL wants a (write|read) first!/) {
-	    require IO::Select;
-	    my $sel = new IO::Select($socket);
-	    my ($timeout) = grep { defined $_ } (@{$arg_hash}{"Timeout","timeout","io_socket_timeout"});
-	    next if (($1 eq 'write') ? $sel->can_write($timeout) : $sel->can_read($timeout));
+	if ( my $cl = $arg_hash->{SSL_cipher_list} ) {
+	    Net::SSLeay::set_cipher_list($ssl, $cl )
+		|| return $socket->error("Failed to set SSL cipher list");
 	}
-	return;
     }
 
+    $ssl ||= ${*$socket}{'_SSL_object'};
+
+    $SSL_ERROR = undef;
+    #DEBUG( 'calling ssleay::accept' );
+    my $rv = Net::SSLeay::accept($ssl);
+    #DEBUG( 'called ssleay::accept rv='.$rv );
+    if ( $rv < 0 ) {
+	unless ( $socket->_set_rw_error( $ssl,$rv )) {
+	    $socket->error("SSL accept attempt failed with unknown error");
+	    delete ${*$self}{'_SSL_opening'};
+    	    ${*$socket}{'_SSL_opened'} = 1;
+	    return $socket->fatal_ssl_error();
+	}
+	return;
+    } elsif ( $rv == 0 ) {
+	$socket->error("SSL connect accept failed because of handshake problems" );
+	delete ${*$self}{'_SSL_opening'};
+	${*$socket}{'_SSL_opened'} = 1;
+	return $socket->fatal_ssl_error();
+    }
+
+    #DEBUG( 'handshake done, socket ready' );
+    # socket opened
+    delete ${*$self}{'_SSL_opening'};
+    ${*$socket}{'_SSL_opened'} = 1;
+
     tie *{$socket}, "IO::Socket::SSL::SSL_HANDLE", $socket;
+
     return $socket;
 }
 
 
 ####### I/O subroutines ########################
+
 sub generic_read {
     my ($self, $read_func, undef, $length, $offset) = @_;
     my $ssl = $self->_get_ssl_object || return;
-
-    my $data = $read_func->($ssl, $length);
-    return $self->error("SSL read error") unless (defined($data) && length($data));
-
     my $buffer=\$_[2];
+    
+    $SSL_ERROR = undef;
+    my $data = $read_func->($ssl, $length);
+    if ( !defined($data)) {
+	$self->_set_rw_error( $ssl,-1 ) || $self->error("SSL read error");
+	return;
+    }
+    
     $length = length($data);
-    $$buffer ||= '';
+    $$buffer = '' if !defined $$buffer;
     $offset ||= 0;
-    if ($offset > length($$buffer)) {
+    if ($offset>length($$buffer)) {
 	$$buffer.="\0" x ($offset-length($$buffer));  #mimic behavior of read
     }
 
@@ -220,8 +360,16 @@ sub generic_read {
 
 sub read {
     my $self = shift;
-    return $self->generic_read($self->blocking ? \&Net::SSLeay::ssl_read_all :
-			       \&Net::SSLeay::read, @_);
+    return $self->generic_read( 
+	$self->blocking ? \&Net::SSLeay::ssl_read_all : \&Net::SSLeay::read, 
+	@_ 
+    );
+}
+
+# contrary to the behavior of read sysread can read partial data
+sub sysread {
+    my $self = shift;
+    return $self->generic_read( \&Net::SSLeay::read, @_ );
 }
 
 sub peek {
@@ -233,46 +381,60 @@ sub peek {
     }
 }
 
-sub write {
-    my ($self, undef, $length, $offset) = @_;
-    my $ssl = $self->_get_ssl_object || return;
 
-    my $buffer = \$_[1];
+sub generic_write {
+    my ($self, $write_all, undef, $length, $offset) = @_;
+
+    my $ssl = $self->_get_ssl_object || return;
+    my $buffer = \$_[2];
+
     my $buf_len = length($$buffer);
     $length ||= $buf_len;
     $offset ||= 0;
     return $self->error("Invalid offset for SSL write") if ($offset>$buf_len);
     return 0 if ($offset == $buf_len);
 
+    $SSL_ERROR = undef;
+    my $written;
+    if ( $write_all ) {
+    	my $data = $length < $buf_len-$offset ? substr($$buffer, $offset, $length) : $$buffer;
+	$written = Net::SSLeay::ssl_write_all($ssl, $data);
+    } else {
+	$written = Net::SSLeay::write_partial( $ssl,$offset,$length,$$buffer );
+    }
+    $written = undef if $written < 0; # Net::SSLeay::write returns -1 not undef on error
+    if ( !defined($written) ) {
+	$self->_set_rw_error( $ssl,-1 )
+	    || $self->error("SSL write error");
+	return;
+    }
 
-    my $written = Net::SSLeay::ssl_write_all
-	($ssl, \substr($$buffer, $offset, $length));
-
-    return $self->error("SSL write error") if (!defined($written) or $written<0);
     return $written;
 }
 
+# if socket is blocking write() should return only on error or
+# if all data are written
+sub write {
+    my $self = shift;
+    return $self->generic_write( scalar($self->blocking),@_ );
+}
+
+# contrary to write syswrite() returns already if only
+# a part of the data is written
+sub syswrite {
+    my $self = shift;
+    return $self->generic_write( 0,@_ );
+}
+
 sub print {
     my $self = shift;
-    my $ssl = $self->_get_ssl_object || return;
-
-    unless ($\ or $,) {
-	foreach my $msg (@_) {
-	    next unless defined $msg;
-	    defined(Net::SSLeay::write($ssl, $msg))
-		|| return $self->error("SSL print error");
-	}
-    } else {
-	defined(Net::SSLeay::write($ssl, join(($, or ''), @_, ($\ or ''))))
-	    || return $self->error("SSL print error");
-    }
-    return 1;
+    my $string = join(($, or ''), @_, ($\ or ''));
+    return $self->write( $string );
 }
 
 sub printf {
     my ($self,$format) = (shift,shift);
-    local $\;
-    return $self->print(sprintf($format, @_));
+    return $self->write(sprintf($format, @_));
 }
 
 sub getc {
@@ -285,10 +447,74 @@ sub readline {
     my $ssl = $self->_get_ssl_object || return;
 
     if (wantarray) {
-	return split (/^/, Net::SSLeay::ssl_read_all($ssl));
+	my ($buf,$err) = Net::SSLeay::ssl_read_all($ssl);
+	return $self->error( "SSL read error" ) if $err;
+	if ( !defined($/) ) {
+	    return $buf;
+	} elsif ( ref($/) ) {
+	    my $size = ${$/};
+	    die "bad value in ref \$/: $size" unless $size>0;
+	    return $buf=~m{\G(.{1,$size})}g;
+	} elsif ( $/ eq '' ) {
+	    return $buf =~m{\G(.*\n\n+|.+)}g;
+	} else {
+	    return $buf =~m{\G(.*$/|.+)}g;
+	}
+    }
+
+    if ( !defined($/) ) {
+	my ($buf,$err) = Net::SSLeay::ssl_read_all($ssl);
+	return $self->error( "SSL read error" ) if $err;
+	return $buf;
+    } elsif ( ref($/) ) {
+	my $size = ${$/};
+	die "bad value in ref \$/: $size" unless $size>0;
+	my ($buf,$err) = Net::SSLeay::ssl_read_all($ssl,$size);
+	return $self->error( "SSL read error" ) if $err;
+	return $buf;
+    } elsif ( $/ ne '' ) {
+    	my $line = Net::SSLeay::ssl_read_until($ssl,$/);
+	return $self->error( "SSL read error" ) if $line eq '';
+	return $line;
+    } else {
+	# $/ is ''
+	# ^.*?\n\n+, need peek to find all \n at the end
+    	die "empty \$/ is not supported if I don't have peek"
+	    if Net::SSLeay::OPENSSL_VERSION_NUMBER() < 0x0090601f;
+
+	# find first occurence of \n\n
+	my $buf = '';
+	my $eon = 0;
+	while (1) { 
+	    defined( Net::SSLeay::peek($ssl,1)) || last; # peek more, can block
+	    my $pending = Net::SSLeay::pending($ssl);
+	    $buf .= Net::SSLeay::peek( $ssl,$pending );  # will not block
+	    if ( !$eon ) {
+	    	my $pos = index( $buf,"\n\n");
+		next if $pos<0; # newlines not found
+		$eon = $pos+2;  # pos after second newline
+	    }
+	    # $eon >= 2  == bytes incl last known \n
+	    while ( index( $buf,"\n",$eon ) == $eon ) {
+	    	# the next char ist \n too
+		$eon++;
+	    }
+	    last if $eon < length($buf); # found last \n before end of buf
+	}
+	if ( $eon > 0 ) {
+	    # found something
+	    # readed peeked data until $eon from $ssl
+	    return Net::SSLeay::ssl_read_all( $ssl,$eon );
+	} else {
+	    # found nothing
+	    # return all what we have
+	    if ( my $l = length($buf)) {
+	    	return Net::SSLeay::ssl_read_all( $ssl,$l );
+	    } else {
+	    	return $self->error( "SSL read error" );
+	    }
+	}
     }
-    my $line = Net::SSLeay::ssl_read_until($ssl);
-    return ($line ne '') ? $line : $self->error("SSL read error");
 }
 
 sub close {
@@ -331,7 +557,8 @@ sub kill_socket {
 
 sub fileno {
     my $self = shift;
-    return ${*$self}{'_SSL_fileno'} || $self->SUPER::fileno();
+    my $fn = ${*$self}{'_SSL_fileno'};
+	return defined($fn) ? $fn : $self->SUPER::fileno();
 }
 
 
@@ -368,13 +595,22 @@ sub start_SSL {
     $socket->configure_SSL($arg_hash) or bless($socket, $original_class) && return;
 
     ${*$socket}{'_SSL_fileno'} = $original_fileno;
-    $arg_hash = ${*$socket}{'_SSL_arguments'};
 
-    my $result = ($arg_hash->{'SSL_server'} 
-		  ? $socket->accept_SSL (${*$socket}{'_SSL_ctx'}, $arg_hash)
-		  : $socket->connect_SSL($socket));
+    my $start_handshake = $arg_hash->{SSL_startHandshake};
+    if ( ! defined($start_handshake) || $start_handshake ) {
+	# if we have no callback force blocking mode
+	#DEBUG( "start handshake" );
+	my $blocking = $socket->blocking(1);
+	my $result = ${*$socket}{'_SSL_arguments'}{SSL_server}
+	    ? $socket->accept_SSL
+	    : $socket->connect_SSL;
+	$socket->blocking(0) if !$blocking;
+    	return $result ? $socket : (bless($socket, $original_class) && ());
+    } else {
+	#DEBUG( "dont start handshake: $socket" );
+    	return $socket; # just return upgraded socket 
+    }
 
-    return $result ? $socket : (bless($socket, $original_class) && ());
 }
 
 sub new_from_fd {
@@ -393,6 +629,7 @@ sub new_from_fd {
     return $class->start_SSL($handle, @_);
 }
 
+
 sub dump_peer_certificate {
     my $ssl = shift()->_get_ssl_object || return;
     return Net::SSLeay::dump_peer_certificate($ssl);
@@ -405,12 +642,16 @@ sub peer_certificate {
     my $cert = ${*$self}{'_SSL_certificate'} ||= Net::SSLeay::get_peer_certificate($ssl) ||
 	return $self->error("Could not retrieve peer certificate");
 
-    my $name = ($field eq "issuer" or $field eq "authority")
-	? Net::SSLeay::X509_get_issuer_name($cert)
-	: Net::SSLeay::X509_get_subject_name($cert);
+    if ($field) {
+	my $name = ($field eq "issuer" or $field eq "authority")
+	    ? Net::SSLeay::X509_get_issuer_name($cert)
+	    : Net::SSLeay::X509_get_subject_name($cert);
 
-    return $self->error("Could not retrieve peer certificate $field") unless ($name);
-    return Net::SSLeay::X509_NAME_oneline($name);
+	return $self->error("Could not retrieve peer certificate $field") unless ($name);
+	return Net::SSLeay::X509_NAME_oneline($name);
+    } else {
+    	return $cert
+    };
 }
 
 sub get_cipher {
@@ -420,7 +661,7 @@ sub get_cipher {
 
 sub errstr {
     my $self = shift;
-    return ((ref($self) ? ${*$self}{'_SSL_last_err'} : $ERROR) or '');
+    return ((ref($self) ? ${*$self}{'_SSL_last_err'} : $SSL_ERROR) or '');
 }
 
 sub fatal_ssl_error {
@@ -441,23 +682,10 @@ sub get_ssleay_error {
 
 sub error {
     my ($self, $error, $destroy_socket) = @_;
-    foreach ($error) {
-	if (/ print / || / write / || / read / || / connect / || / accept /) {
-	    my $ssl = ${*$self}{'_SSL_object'};
-	    my $ssl_error = $ssl ? Net::SSLeay::get_error($ssl, -1) : 0;
-	    if ($ssl_error == Net::SSLeay::ERROR_WANT_READ()) {
-		$error.="\nSSL wants a read first!";
-	    } elsif ($ssl_error == Net::SSLeay::ERROR_WANT_WRITE()) {
-		$error.="\nSSL wants a write first!";
-	    } else {
-		$error.=Net::SSLeay::ERR_error_string
-		    (Net::SSLeay::ERR_get_error());
-	    }
-	}
-    }
+    $error .= Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error());
     carp $error."\n".$self->get_ssleay_error() if $DEBUG;
-    ${*$self}{'_SSL_last_err'} = $error if (ref($self));
-    $ERROR = $error;
+    $SSL_ERROR = dualvar( -1, $error );
+    ${*$self}{'_SSL_last_err'} = $SSL_ERROR if (ref($self));
     return;
 }
 
@@ -472,8 +700,7 @@ sub DESTROY {
 #######Extra Backwards Compatibility Functionality#######
 sub socket_to_SSL { IO::Socket::SSL->start_SSL(@_); }
 sub socketToSSL { IO::Socket::SSL->start_SSL(@_); }
-sub sysread { &IO::Socket::SSL::read; }
-sub syswrite { &IO::Socket::SSL::write; }
+
 sub issuer_name { return(shift()->peer_certificate("issuer")) }
 sub subject_name { return(shift()->peer_certificate("subject")) }
 sub get_peer_certificate { return shift() }
@@ -492,15 +719,13 @@ sub opened {
     return IO::Handle::opened($self) && ${*$self}{'_SSL_opened'};
 }
 
-sub want_read {
+sub opening {
     my $self = shift;
-    return scalar($self->errstr() =~ /SSL wants a read first!/);
+    return ${*$self}{'_SSL_opening'};
 }
 
-sub want_write {
-    my $self = shift;
-    return scalar($self->errstr() =~ /SSL wants a write first!/);
-}
+sub want_read  { shift->errstr == SSL_WANT_READ }
+sub want_write { shift->errstr == SSL_WANT_WRITE }
 
 
 #Redundant IO::Handle functionality
@@ -523,6 +748,7 @@ sub recv   { croak("Use of recv() not implemented in IO::Socket::SSL; use read/s
 package IO::Socket::SSL::SSL_HANDLE;
 use strict;
 use vars qw($HAVE_WEAKREF);
+use Errno 'EBADF';
 
 BEGIN {
     local ($@, $SIG{__DIE__});
@@ -539,16 +765,19 @@ sub TIEHANDLE {
     bless \$handle, $class;
 }
 
-sub READ     { ${shift()}->read     (@_) }
+sub READ     { ${shift()}->sysread  (@_) }
 sub READLINE { ${shift()}->readline (@_) }
 sub GETC     { ${shift()}->getc     (@_) }
 
 sub PRINT    { ${shift()}->print    (@_) }
 sub PRINTF   { ${shift()}->printf   (@_) }
-sub WRITE    { ${shift()}->write    (@_) }
+sub WRITE    { ${shift()}->syswrite (@_) }
 
 sub FILENO   { ${shift()}->fileno   (@_) }
 
+sub TELL     { $! = EBADF; return -1 }
+sub BINMODE  { return 0 }  # not perfect, but better than not implementing the method
+
 sub CLOSE {                          #<---- Do not change this function!
     my $ssl = ${$_[0]};
     local @_;
@@ -559,11 +788,14 @@ sub CLOSE {                          #<---- Do not change this function!
 package IO::Socket::SSL::SSL_Context;
 use strict;
 
+# should be better taken from Net::SSLeay, but they are not (yet) defined there
+use constant SSL_MODE_ENABLE_PARTIAL_WRITE => 1;
+use constant SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER => 2;
+
 # Note that the final object will actually be a reference to the scalar
 # (C-style pointer) returned by Net::SSLeay::CTX_*_new() so that
 # it can be blessed.
-sub new
-{
+sub new {
     my $class = shift;
     my $arg_hash = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
 
@@ -582,13 +814,21 @@ sub new
 	$ctx = /^sslv2$/i ? Net::SSLeay::CTX_v2_new()    :
 	       /^sslv3$/i ? Net::SSLeay::CTX_v3_new()    :
 	       /^tlsv1$/i ? Net::SSLeay::CTX_tlsv1_new() :
-	                    Net::SSLeay::CTX_new();
+			    Net::SSLeay::CTX_new();
     }
 
     $ctx || return IO::Socket::SSL->error("SSL Context init failed");
 
     Net::SSLeay::CTX_set_options($ctx, Net::SSLeay::OP_ALL());
 
+    # SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER makes syswrite return if at least one
+    # buffer was written and not block for the rest
+    # SSL_MODE_ENABLE_PARTIAL_WRITE can be necessary for non-blocking because we
+    # cannot guarantee, that the location of the buffer stays constant
+    Net::SSLeay::CTX_set_mode( $ctx, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER
+    	|SSL_MODE_ENABLE_PARTIAL_WRITE);
+
+
     my ($verify_mode, $verify_cb) = @{$arg_hash}{'SSL_verify_mode','SSL_verify_callback'};
     unless ($verify_mode == Net::SSLeay::VERIFY_NONE())
     {
@@ -615,13 +855,46 @@ sub new
 	    Net::SSLeay::CTX_set_default_passwd_cb($ctx, $arg_hash->{'SSL_passwd_cb'});
 	}
 
-	Net::SSLeay::CTX_use_PrivateKey_file
-	    ($ctx, $arg_hash->{'SSL_key_file'}, $filetype)
-	    || return IO::Socket::SSL->error("Failed to open Private Key");
+	if ( my $pkey= $arg_hash->{SSL_key} ) {
+	    # binary, e.g. EVP_PKEY*
+	    Net::SSLeay::CTX_use_PrivateKey($ctx, $pkey)
+		|| return IO::Socket::SSL->error("Failed to use Private Key");
+	} elsif ( my $f = $arg_hash->{SSL_key_file} ) {
+	    Net::SSLeay::CTX_use_PrivateKey_file($ctx, $f, $filetype)
+		|| return IO::Socket::SSL->error("Failed to open Private Key");
+	}
 
-	Net::SSLeay::CTX_use_certificate_chain_file
-	    ($ctx, $arg_hash->{'SSL_cert_file'})
-	    || return IO::Socket::SSL->error("Failed to open Certificate");
+	if ( my $x509 = $arg_hash->{SSL_cert} ) {
+	    # binary, e.g. X509*
+	    # we habe either a single certificate or a list with
+	    # a chain of certificates
+	    my @x509 = ref($x509) eq 'ARRAY' ? @$x509: ($x509);
+	    my $cert = shift @x509;
+	    Net::SSLeay::CTX_use_certificate( $ctx,$cert ) 
+	    	|| return IO::Socket::SSL->error("Failed to use Certificate");
+	    foreach my $ca (@x509) {
+	    	Net::SSLeay::CTX_add_extra_chain_cert( $ctx,$ca ) 
+	    	    || return IO::Socket::SSL->error("Failed to use Certificate");
+	    }
+	} elsif ( my $f = $arg_hash->{SSL_cert_file} ) {
+	    Net::SSLeay::CTX_use_certificate_chain_file($ctx, $f)
+		|| return IO::Socket::SSL->error("Failed to open Certificate");
+	}
+
+	if ( my $dh = $arg_hash->{SSL_dh} ) {
+	    # binary, e.g. DH*
+	    Net::SSLeay::CTX_set_tmp_dh( $ctx,$dh )
+	    	|| return IO::Socket::SSL->error( "Failed to set DH from SSL_dh" );
+	} elsif ( my $f = $arg_hash->{SSL_dh_file} ) {
+	    my $bio = Net::SSLeay::BIO_new_file( $f,'r' ) 
+	    	|| return IO::Socket::SSL->error( "Failed to open DH file $f" );
+	    my $dh = Net::SSLeay::PEM_read_bio_DHparams($bio);
+	    Net::SSLeay::BIO_free($bio);
+	    $dh || return IO::Socket::SSL->error( "Failed to read PEM for DH from $f - wrong format?" );
+	    my $rv = Net::SSLeay::CTX_set_tmp_dh( $ctx,$dh );
+	    Net::SSLeay::DH_free( $dh );
+	    $rv || return IO::Socket::SSL->error( "Failed to set DH from $f" );
+	}
     }
 
     my $verify_callback = $verify_cb &&
@@ -646,7 +919,7 @@ sub new
 	    return IO::Socket::SSL->error("Session caches not supported for Net::SSLeay < v1.26");
 	} else {
 	    $ctx_object->{'session_cache'} =
-		new IO::Socket::SSL::Session_Cache($arg_hash) || undef;
+		IO::Socket::SSL::Session_Cache->new($arg_hash) || undef;
 	}
     }
 
@@ -663,7 +936,7 @@ sub session_cache {
     my $session = shift;
 
     return (defined($session) ? $cache->add_session($key, $session)
-	                      : $cache->get_session($key));
+			      : $cache->get_session($key));
 }
 
 sub has_session_cache {
@@ -739,7 +1012,7 @@ sub DESTROY {
 }
 
 
-'True Value';
+1;
 
 
 =head1 NAME
@@ -750,15 +1023,15 @@ IO::Socket::SSL -- Nearly transparent SSL encapsulation for IO::Socket::INET.
 
     use IO::Socket::SSL;
 
-    my $client = new IO::Socket::SSL("www.example.com:https");
+    my $client = IO::Socket::SSL->new("www.example.com:https");
 
-    if (defined $client) {
-        print $client "GET / HTTP/1.0\r\n\r\n";
-        print <$client>;
-        close $client;
+    if ($client) {
+	print $client "GET / HTTP/1.0\r\n\r\n";
+	print <$client>;
+	close $client;
     } else {
-        warn "I encountered a problem: ",
-          IO::Socket::SSL::errstr();
+	warn "I encountered a problem: ",
+	  IO::Socket::SSL::errstr();
     }
 
 
@@ -778,6 +1051,8 @@ If you have used this module before, read on, as versions 0.93 and above
 have several changes from the previous IO::Socket::SSL versions (especially
 see the note about return values).
 
+If you are using non-blocking sockets read on, as version 0.98 added better
+support for non-blocking.
 
 =head1 METHODS
 
@@ -803,9 +1078,12 @@ which auto-negotiates between SSLv2 and SSLv3.  You may specify 'SSLv2', 'SSLv3'
 
 =item SSL_cipher_list
 
-If you do not care for the default list of ciphers ('ALL:!LOW:!EXP'), then look in
-the OpenSSL documentation (L<http://www.openssl.org/docs/apps/ciphers.html#CIPHER_STRINGS>),
-and specify a different set with this option.
+If this option is set the cipher list for the connection will be set to the
+given value, e.g. something like 'ALL:!LOW:!EXP:!ADH'. Look into the OpenSSL 
+documentation (L<http://www.openssl.org/docs/apps/ciphers.html#CIPHER_STRINGS>)
+for more details.
+If this option is not used the openssl builtin default is used which is suitable
+for most cases.
 
 =item SSL_use_cert
 
@@ -821,6 +1099,13 @@ specify a different location.  Keys should be PEM formatted, and if they are
 encrypted, you will be prompted to enter a password before the socket is formed
 (unless you specified the SSL_passwd_cb option).
 
+=item SSL_key
+
+This is an EVP_PKEY* and can be used instead of SSL_key_file.
+Useful if you don't have your key in a file but create it dynamically or get it from
+a string (see openssl PEM_read_bio_PrivateKey etc for getting a EVP_PKEY* from
+a string).
+
 =item SSL_cert_file
 
 If your SSL certificate is not in the default place (F<certs/server-cert.pem> for servers,
@@ -829,6 +1114,23 @@ location of your certificate.  Note that a key and certificate are only required
 SSL server, so you do not need to bother with these trifling options should you be
 setting up an unauthenticated client.
 
+=item SSL_cert
+
+This is an X509* or an array of X509*.
+The first X509* is the internal representation of the certificate while the following
+ones are extra certificates. Useful if you create your certificate dynamically (like
+in a SSL intercepting proxy) or get it from a string (see openssl PEM_read_bio_X509 etc
+for getting a X509* from a string).
+
+=item SSL_dh_file
+
+If you want Diffie-Hellman key exchange you need to supply a suitable file here
+or use the SSL_dh parameter. See dhparam command in openssl for more information.
+
+=item SSL_dh
+
+Like SSL_dh_file, but instead of giving a file you use a preloaded or generated DH*.
+
 =item SSL_passwd_cb
 
 If your private key is encrypted, you might not want the default password prompt from
@@ -898,11 +1200,7 @@ positive cache size.  For successive connections, pass the SSL_reuse_ctx option
 the new() calls (or use set_default_context()) to make use of the cached sessions.
 The session cache size refers to the number of unique host/port pairs that can be
 stored at one time; the oldest sessions in the cache will be removed if new ones are
-added.  This requires an unofficial patched version of Net::SSLeay to work; check
-the patches directory, or download the altered version at L<http://www.fas.harvard.edu/~behrooz/Net_SSLeay.pm-1.26.tar.gz>.
-I have contacted Sampo Kellomki about the patch, and he has assured me that it
-will appear in the next official release, but that has (as of this writing) not
-yet seen the light of CPAN.
+added.  
 
 =item SSL_error_trap
 
@@ -974,15 +1272,18 @@ only fields it can return are "authority" and "owner" (or "issuer" and "subject"
 you want to use OpenSSL names), corresponding to the certificate authority that signed the
 peer certificate and the owner of the peer certificate.  This function returns a string
 with all the information about the particular field in one parsable line.
+If no field is given it returns the full certificate (x509).
 
 =item B<errstr()>
 
 Returns the last error (in string form) that occurred.  If you do not have a real
 object to perform this method on, call IO::Socket::SSL::errstr() instead.
+
 For read and write errors on non-blocking sockets, this method may include the string
 C<SSL wants a read first!> or C<SSL wants a write first!> meaning that the other side
 is expecting to read from or write to the socket and wants to be satisfied before you
-get to do anything.
+get to do anything. But with version 0.98 you are better comparing the global exported 
+variable $SSL_ERROR against the exported symbols SSL_WANT_READ and SSL_WANT_WRITE.
 
 =item B<< IO::Socket::SSL->start_SSL($socket, ... ) >>
 
@@ -992,7 +1293,12 @@ a call to new().  If you are using this function on an accept()ed socket, you mu
 set the parameter "SSL_server" to 1, i.e. IO::Socket::SSL->start_SSL($socket, SSL_server => 1).
 If you have a class that inherits from IO::Socket::SSL and you want the $socket to be blessed
 into your own class instead, use MyClass->start_SSL($socket) to achieve the desired effect.
-Note that if start_SSL() fails in SSL negotiation, $socket will remain blessed in its original class.
+
+Note that if start_SSL() fails in SSL negotiation, $socket will remain blessed in its 
+original class.  For non-blocking sockets you better just upgrade the socket to 
+IO::Socket::SSL and call accept_SSL or connect_SSL and the upgraded object. To
+just upgrade the socket set B<SSL_startHandshake> explicitly to 0. If you call start_SSL
+w/o this parameter it will revert to blocking behavior for accept_SSL and connect_SSL.
 
 =item B<< IO::Socket::SSL->new_from_fd($fd, ...) >>
 
@@ -1045,7 +1351,7 @@ A few changes have gone into IO::Socket::SSL v0.93 and later with respect to
 return values.  The behavior on success remains unchanged, but for I<all> functions,
 the return value on error is now an empty list.  Therefore, the return value will be
 false in all contexts, but those who have been using the return values as arguments
-to subroutines (like C<mysub(new IO::Socket::SSL(...), ...)>) may run into problems.
+to subroutines (like C<mysub(IO::Socket::SSL(...)->new, ...)>) may run into problems.
 The moral of the story: I<always> check the return values of these functions before
 using them in any way that you consider meaningful.
 
@@ -1105,18 +1411,13 @@ See the 'example' directory.
 
 =head1 BUGS
 
-I have never shipped a module with a known bug, and IO::Socket::SSL is no
-different.  If you feel that you have found a bug in the module and you are
-using the latest versions of Net::SSLeay and OpenSSL, send an email immediately to
-<behrooz at fas.harvard.edu> with a subject of 'IO::Socket::SSL Bug'.  Until very
-recently, I did not receive any indication when a bug was submitted to the CPAN
-request tracker, so e-mail is a much more reliable tool.  I am
-I<not responsible> for problems in your code, so make sure that an example
-actually works before sending it. It is merely acceptable if you send me a bug
-report, it is better if you send a small chunk of code that points it out,
-and it is best if you send a patch--if the patch is good, you might see a release the
-next day on CPAN. Otherwise, it could take weeks . . .
+IO::Socket::SSL is not threadsafe.
+This is because IO::Socket::SSL is based on Net::SSLeay which 
+uses a global object to access some of the API of openssl
+and is therefore not threadsafe.
 
+IO::Socket::SSL does not work together with Storable::fd_retrieve/fd_store.
+See BUGS file for more information and how to work around the problem.
 
 =head1 LIMITATIONS
 
@@ -1138,18 +1439,27 @@ The following functions are deprecated and are only retained for compatibility:
 
 =item context_init()
 
-(use the SSL_reuse_ctx option if you want to re-use a context)
+use the SSL_reuse_ctx option if you want to re-use a context
 
 
 =item socketToSSL() and socket_to_SSL()
 
-(use IO::Socket::SSL->start_SSL() instead)
+use IO::Socket::SSL->start_SSL() instead
 
 
-=item get_peer_certificate() and friends
+=item get_peer_certificate()
 
-(use the peer_certificate() function instead)
+use the peer_certificate() function instead.
+Used to return X509_Certificate with methods subject_name and issuer_name.
+Now simply returns $self which has these methods (although depreceated).
 
+=item issuer_name()
+
+use peer_certificate( 'issuer' ) instead
+
+=item subject_name()
+
+use peer_certificate( 'subject' ) instead
 
 =back
 
@@ -1173,12 +1483,18 @@ IO::Socket::INET, IO::Socket::INET6, Net::SSLeay.
 
 =head1 AUTHORS
 
+Steffen Ullrich, <steffen at genua.de> is the current maintainer.
+
 Peter Behroozi, <behrooz at fas.harvard.edu> (Note the lack of an "i" at the end of "behrooz")
 
 Marko Asplund, <marko.asplund at kronodoc.fi>, was the original author of IO::Socket::SSL.
 
+Patches incorporated from various people, see file Changes.
+
 =head1 COPYRIGHT
 
+Working support for non-blocking was added by Steffen Ullrich.
+
 The rewrite of this module is Copyright (C) 2002-2005 Peter Behroozi.
 
 The original versions of this module are Copyright (C) 1999-2002 Marko Asplund.
